package com.annahid.games.example.scenes;

import com.annahid.libs.artenus.core.Scene;
import com.annahid.libs.artenus.core.Stage;
import com.annahid.libs.artenus.data.Point2D;
import com.annahid.libs.artenus.entities.EntityCollection;
import com.annahid.libs.artenus.graphics.animation.Trends;
import com.annahid.libs.artenus.graphics.TextureManager;
import com.annahid.games.example.R;
import com.annahid.libs.artenus.graphics.animation.ImageAnimation;
import com.annahid.libs.artenus.graphics.sprites.ImageSprite;
import com.annahid.libs.artenus.input.GameInput;
import com.annahid.libs.artenus.input.InputListener;
import com.annahid.libs.artenus.input.SlideInput;
import com.annahid.libs.artenus.sound.SoundManager;

/**
 * This is the game scene, which is displayed if you click the play button.
 */
public class GameScene extends Scene implements InputListener {
    /**
     * Yes, our main and only character is a girl! Attributed to the game Give Me a Kiss!
     */
    private ImageSprite girl;

    /**
     * This is the HUD for the direction knob. The knob spawns wherever you first touch the screen.
     */
    private ImageSprite arrowSprite;

    /**
     * Holds movement direction of the character.
     */
    private Point2D direction = new Point2D(0, 0);

    /**
     * Walking animation handler for our character.
     */
    private final ImageAnimation walkingAnimation;

    /**
     * This variable is used to avoid multiple playing of the step sound.
     */
    private int savedFrame = 0;

    /**
     * Here is an example of entity grouping. Your game scene has two layers of entities.
     * The Game layer contains all game objects.
     */
    private EntityCollection gameLayer = new EntityCollection();

    /**
     * The HUD layer contains, well, the HUD!
     */
    private EntityCollection hudLayer = new EntityCollection();

    /**
     * Animation handlers are pretty safe to instantiate in the constructor. Sprites are not!
     *
     * @param parentStage parent stage
     */
    public GameScene(Stage parentStage) {
        super(parentStage);
        walkingAnimation = new ImageAnimation(
                new int[]{0, 1, 2, 1, 0, 3, 4, 3},
                Trends.LOOP,
                0
        );
    }

    /**
     * Override this method whenever you have texture that are used only in one scene.
     * This helps you remove bloat from your game's global scope. Local textures are unloaded
     * automatically when the scene changes.
     */
    @Override
    public void onLocalLoad() {
        TextureManager.addLocal(R.raw.anim_girl, R.raw.arrow);
    }

    @Override
    public void onLoaded() {
        // Setup the atlas cutout
        ImageSprite.Cutout cutout = new ImageSprite.Cutout(150, 120, 6, 2);

        // Create the character
        girl = new ImageSprite(R.raw.anim_girl, cutout);
        girl.setPosition(stage.getLogicalWidth() / 2, stage.getLogicalHeight() / 2);
        gameLayer.add(girl);

        // Create the HUD entity
        arrowSprite = new ImageSprite(R.raw.arrow, new ImageSprite.Cutout(256, 256, 1));
        arrowSprite.setAlpha(0);
        arrowSprite.setPosition(0, 0);
        hudLayer.add(arrowSprite);

        // You could have done this before adding items to them, it wouldn't make a difference.
        // It is only important to add the HUD layer after the game layer, so its contents are
        // displayed above game contents.
        add(gameLayer);
        add(hudLayer);

        // The input controller is added to the scene just like any other entity. It is recommended
        // to add it directly to the scene, and not in a nested collection.
        SlideInput controller = new SlideInput();
        controller.setEffectiveArea(0, 0, stage.getLogicalWidth() - 300, 600);
        controller.setListener(this);
        add(controller);

        super.onLoaded();
    }

    /**
     * Called by the framework for each frame, to advance the animation.
     * ALWAYS CALL THE SUPERCLASS METHOD. It doesn't really matter where, but I usually call it
     * first thing so all internal animation handling is completed before I begin.
     *
     * @param elapsedTime The time elapsed since last frame
     */
    @Override
    public void advance(float elapsedTime) {
        super.advance(elapsedTime);
        float dx = direction.x * elapsedTime * 100;
        float dy = direction.y * elapsedTime * 100;
        girl.move(dx, dy);
        checkStepSound();
    }

    /**
     * This is a very simple example of how you can accommodate sound in your game.
     */
    private void checkStepSound() {
        final int frame = girl.getCurrentFrame();

        if (savedFrame != frame && (frame == 1 || frame == 3)) {
            SoundManager.playSound(R.raw.step);
            savedFrame = frame;
        }
    }

    /**
     * This method is called on the input listener whenever the input status changes.
     *
     * @param input The calling input manager. Use this to retrieve information about the event
     */
    @Override
    public void inputStatusChanged(GameInput input) {
        SlideInput slideInput = (SlideInput) input;
        direction = input.getDirection();

        if (direction.x == 0 && direction.y == 0) {
            // If there is no direction, freeze everything
            arrowSprite.setAlpha(0);
            girl.setAnimation(null);
        } else {
            // Otherwise, show the direction HUD entity, and rotate the character
            arrowSprite.setAlpha(0.5f);
            arrowSprite.setPosition(slideInput.getRefX(), slideInput.getRefY());
            arrowSprite.setRotation((float) Math.toDegrees(Math.atan2(direction.y, direction.x)));
            girl.setAnimation(walkingAnimation);
            girl.setRotation(90 + arrowSprite.getRotation());
        }
    }

    /**
     * In internal scenes, you can show confirmation dialogs here to ask the user if they are
     * sure to exit. If you like your users, never do that in the starting scene! If a user hits
     * the back button on the game from the start screen, he most probably means it. There is no
     * need to ask!
     *
     * @return true
     */
    @Override
    public boolean onBackButton() {
        stage.setScene(new MenuScene(stage));
        return true;
    }
}
